94057c3d3e
PR Test (NPU) / check-changes (push) Has been cancelled
PR Test (NPU) / pr-gate (push) Has been cancelled
PR Test (NPU) / set-image-config (push) Has been cancelled
PR Test (NPU) / stage-b-test-1-npu-a2 (0) (push) Has been cancelled
PR Test (NPU) / stage-b-test-1-npu-a2 (1) (push) Has been cancelled
PR Test (NPU) / stage-b-test-2-npu-a2 (0) (push) Has been cancelled
PR Test (NPU) / stage-b-test-2-npu-a2 (1) (push) Has been cancelled
PR Test (NPU) / stage-b-test-4-npu-a3 (push) Has been cancelled
PR Test (NPU) / stage-b-test-16-npu-a3 (push) Has been cancelled
PR Test (NPU) / multimodal-gen-test-1-npu-a3 (push) Has been cancelled
PR Test (NPU) / multimodal-gen-test-2-npu-a3 (push) Has been cancelled
PR Test (Arm64) / pr-gate (push) Has been cancelled
PR Test (Arm64) / check-changes (push) Has been cancelled
PR Test (Arm64) / build-test (push) Has been cancelled
PR Test (sgl-router) / gate (push) Has been cancelled
PR Test (sgl-router) / tier-1 — lint (push) Has been cancelled
PR Test (sgl-router) / tier-2 — build + test (push) Has been cancelled
PR Test (sgl-router) / tier-3 — docker (placeholder) (push) Has been cancelled
PR Test (sgl-router) / tier-3 — k8s integration (push) Has been cancelled
PR Test (sgl-router) / tier-3 — e2e (push) Has been cancelled
PR Test (sgl-router) / finish (push) Has been cancelled
PR Test (NPU) / single-node-poc (map[name:qwen3_6_27b_w8a8_1p_in64k_out1k_50ms runner:linux-aarch64-a3-2 test_case:test/registered/ascend/performance/qwen3_6_27b/test_npu_qwen3_6_27b_w8a8_1p_in64k_out1k_50ms.py test_type:perf]) (push) Has been cancelled
PR Test (NPU) / pr-test-npu-finish (push) Has been cancelled
PR Test (Xeon) / pr-gate (push) Has been cancelled
PR Test (Xeon) / check-changes (push) Has been cancelled
PR Test (Xeon) / build-test (, xeon-gnr, base-b-test-cpu) (push) Has been cancelled
PR Test (XPU) / check-changes (push) Has been cancelled
PR Test (XPU) / pr-gate (push) Has been cancelled
PR Test (XPU) / stage-a-test-1-gpu-xpu (push) Has been cancelled
PR Test (XPU) / wait-for-stage-a (push) Has been cancelled
PR Test (XPU) / stage-b-test-1-gpu-xpu (push) Has been cancelled
PR Test (XPU) / finish (push) Has been cancelled
CI Model Inventory / build-inventory (push) Has been cancelled
Lint / lint (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark Compilation Check (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark - Manual Policy (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark - Request Processing (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark Summary (push) Has been cancelled
PR Test (SMG) / build-wheel (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on windows (x86_64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on macos (x86_64 - auto) (push) Has been cancelled
PR Test (SMG) / python-unit-tests (push) Has been cancelled
PR Test (SMG) / unit-tests (push) Has been cancelled
PR Test (SMG) / benchmarks (push) Has been cancelled
PR Test (SMG) / chat-completions (push) Has been cancelled
PR Test (SMG) / chat-completions-4gpu (push) Has been cancelled
PR Test (SMG) / e2e (push) Has been cancelled
PR Test (SMG) / docker-build-test (push) Has been cancelled
PR Test (SMG) / k8s-integration (push) Has been cancelled
PR Test (SMG) / finish (push) Has been cancelled
PR Test (SMG) / summarize-benchmarks (push) Has been cancelled
Release SGLang Model Gateway Docker Image / publish (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on macos (aarch64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (aarch64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (x86_64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (aarch64 - musllinux_1_1) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (x86_64 - musllinux_1_1) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / Build SDist (push) Has been cancelled
Release SGLang Model Gateway to PyPI / Upload to PyPI (push) Has been cancelled
Release SGLang Kernels / build-cu129-matrix (aarch64, 12.9, 3.10, arm-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / build-cu129-matrix (x86_64, 12.9, 3.10, x64-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / release-cu129 (push) Has been cancelled
Release SGLang Kernels / build-cu130-matrix (aarch64, 13.0, 3.10, arm-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / build-cu130-matrix (x86_64, 13.0, 3.10, x64-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / release-cu130 (push) Has been cancelled
Release SGLang Kernels / build-rocm-matrix (3.10, 700) (push) Has been cancelled
Release SGLang Kernels / build-rocm-matrix (3.10, 720) (push) Has been cancelled
Release SGLang Kernels / release-rocm700 (push) Has been cancelled
Release SGLang Kernels / release-rocm720 (push) Has been cancelled
Release SGLang Kernels / build-musa43 (43, 3.10) (push) Has been cancelled
Release SGLang Kernels / release-musa43 (push) Has been cancelled
239 lines
7.6 KiB
Python
239 lines
7.6 KiB
Python
from typing import Optional
|
|
|
|
import torch
|
|
import triton
|
|
import triton.language as tl
|
|
|
|
from sglang.kernels.ops.gemm.lora_tuning_config import get_lora_expand_config
|
|
from sglang.srt.lora.utils import LoRABatchInfo
|
|
from sglang.srt.utils import cached_triton_kernel
|
|
|
|
|
|
@cached_triton_kernel(
|
|
lambda _, kwargs: (kwargs["NUM_SLICES"], kwargs["BLOCK_M"], kwargs["OUTPUT_DIM"])
|
|
)
|
|
@triton.jit(do_not_specialize=["num_segs", "output_stride_0", "output_stride_1"])
|
|
def _chunked_lora_expand_kernel(
|
|
# Pointers to matrices
|
|
x,
|
|
weights,
|
|
output,
|
|
# Output strides may differ from OUTPUT_DIM when compact LoRA output is
|
|
# accumulated into a wider base projection.
|
|
output_stride_0,
|
|
output_stride_1,
|
|
# Information on sequence lengths and weight id
|
|
seg_indptr,
|
|
weight_indices,
|
|
lora_ranks,
|
|
permutation,
|
|
num_segs,
|
|
# For fused output scaling
|
|
scalings,
|
|
# Offsets of q/k/v slice on output dimension
|
|
slice_offsets,
|
|
# Meta parameters
|
|
NUM_SLICES: tl.constexpr,
|
|
OUTPUT_DIM: tl.constexpr,
|
|
MAX_RANK: tl.constexpr, # K = R
|
|
BLOCK_M: tl.constexpr,
|
|
BLOCK_N: tl.constexpr,
|
|
BLOCK_K: tl.constexpr,
|
|
):
|
|
"""
|
|
Computes a chunked SGMV for LoRA expand operations.
|
|
|
|
When a sequence's rank is 0, the kernel is essentially a no-op, following
|
|
the convention in pytorch where the product of two matrices of shape (m, 0)
|
|
and (0, n) is an all-zero matrix of shape (m, n).
|
|
|
|
Args:
|
|
x (Tensor): The input tensor, which is the result of the LoRA A projection.
|
|
Shape: (s, num_slices * K), where s is the sum of all sequence lengths in the
|
|
batch and K is the maximum LoRA rank.
|
|
weights (Tensor): The LoRA B weights for all adapters.
|
|
Shape: (num_lora, output_dim, K).
|
|
output (Tensor): The output tensor where the result is stored.
|
|
Shape: (s, output_dim) or a wider base output.
|
|
"""
|
|
x_stride_0: tl.constexpr = NUM_SLICES * MAX_RANK
|
|
x_stride_1: tl.constexpr = 1
|
|
|
|
w_stride_0: tl.constexpr = OUTPUT_DIM * MAX_RANK
|
|
w_stride_1: tl.constexpr = MAX_RANK
|
|
w_stride_2: tl.constexpr = 1
|
|
|
|
pid_s = tl.program_id(axis=2)
|
|
if pid_s >= num_segs:
|
|
return
|
|
|
|
seg_start = tl.load(seg_indptr + pid_s)
|
|
seg_end = tl.load(seg_indptr + pid_s + 1)
|
|
if seg_start == seg_end:
|
|
return
|
|
|
|
# Current block computes sequence with batch_id,
|
|
# which starts from row seg_start of x with length seg_len.
|
|
# qkv_id decides which of q,k,v to compute (0: q, 1: k, 2: v)
|
|
w_index = tl.load(weight_indices + pid_s)
|
|
cur_rank = tl.load(lora_ranks + w_index)
|
|
|
|
# If rank is 0, this kernel is a no-op.
|
|
if cur_rank == 0:
|
|
return
|
|
|
|
slice_id = tl.program_id(axis=1)
|
|
slice_start = tl.load(slice_offsets + slice_id)
|
|
slice_end = tl.load(slice_offsets + slice_id + 1)
|
|
|
|
scaling = tl.load(scalings + w_index)
|
|
# Adjust K (rank) according to the specific LoRA adapter
|
|
cur_rank = tl.minimum(MAX_RANK, cur_rank)
|
|
|
|
# Map logical sequence index to physical index
|
|
s_offset_logical = tl.arange(0, BLOCK_M) + seg_start
|
|
s_offset_physical = tl.load(
|
|
permutation + s_offset_logical, mask=s_offset_logical < seg_end, other=0
|
|
)
|
|
|
|
# Create pointers for the first block of x and weights[batch_id][n_start: n_end][:]
|
|
# The pointers will be advanced as we move in the K direction
|
|
# and accumulate
|
|
pid_n = tl.program_id(axis=0)
|
|
n_offset = tl.arange(0, BLOCK_N) + pid_n * BLOCK_N + slice_start
|
|
k_offset = tl.arange(0, BLOCK_K)
|
|
|
|
x_ptrs = (
|
|
x
|
|
+ slice_id * cur_rank * x_stride_1
|
|
+ (s_offset_physical[:, None] * x_stride_0 + k_offset[None, :] * x_stride_1)
|
|
)
|
|
w_ptrs = (weights + w_index * w_stride_0) + (
|
|
k_offset[:, None] * w_stride_2 + n_offset[None, :] * w_stride_1
|
|
)
|
|
|
|
# Iterate to compute the block in output matrix
|
|
partial_sum = tl.zeros((BLOCK_M, BLOCK_N), dtype=tl.float32)
|
|
for k in range(0, tl.cdiv(cur_rank, BLOCK_K)):
|
|
x_tile = tl.load(
|
|
x_ptrs,
|
|
mask=(s_offset_logical[:, None] < seg_end)
|
|
& (k_offset[None, :] < cur_rank - k * BLOCK_K),
|
|
other=0.0,
|
|
)
|
|
w_tile = tl.load(
|
|
w_ptrs,
|
|
mask=(k_offset[:, None] < cur_rank - k * BLOCK_K)
|
|
& (n_offset[None, :] < slice_end),
|
|
other=0.0,
|
|
)
|
|
partial_sum += tl.dot(x_tile, w_tile)
|
|
|
|
x_ptrs += BLOCK_K * x_stride_1
|
|
w_ptrs += BLOCK_K * w_stride_2
|
|
|
|
# Store result to output matrix
|
|
partial_sum *= scaling
|
|
partial_sum = partial_sum.to(x.dtype.element_ty)
|
|
output_ptr = output + (
|
|
s_offset_physical[:, None] * output_stride_0
|
|
+ n_offset[None, :] * output_stride_1
|
|
)
|
|
output_mask = (s_offset_logical[:, None] < seg_end) & (
|
|
n_offset[None, :] < slice_end
|
|
)
|
|
partial_sum += tl.load(output_ptr, mask=output_mask, other=0.0)
|
|
tl.store(output_ptr, partial_sum, mask=output_mask)
|
|
|
|
|
|
def chunked_sgmv_lora_expand_forward(
|
|
x: torch.Tensor,
|
|
weights: torch.Tensor,
|
|
batch_info: LoRABatchInfo,
|
|
slice_offsets: torch.Tensor,
|
|
max_slice_size: int,
|
|
base_output: Optional[torch.Tensor],
|
|
) -> torch.Tensor:
|
|
|
|
# x: (s, slice_num * r)
|
|
# weights: (num_lora, output_dim, r)
|
|
# slice_offsets: boundaries for different slices in the output dimension
|
|
# output: (s, output_dim)
|
|
|
|
# Compute lora_output with shape (s, output_dim) as follows:
|
|
# For each slice i, accumulates:
|
|
# lora_output[:, slice_offsets[i]:slice_offsets[i+1]] += scaling * sgemm(x[:, i*cur_rank:(i+1)*cur_rank], weights[:, slice_offsets[i]:slice_offsets[i+1], :])
|
|
|
|
assert x.is_contiguous()
|
|
assert weights.is_contiguous()
|
|
assert len(x.shape) == 2
|
|
assert len(weights.shape) == 3
|
|
|
|
# Get dims
|
|
M = x.shape[0]
|
|
input_dim = x.shape[1]
|
|
OUTPUT_DIM = weights.shape[1]
|
|
MAX_RANK = weights.shape[2]
|
|
num_slices = len(slice_offsets) - 1
|
|
assert input_dim == num_slices * MAX_RANK
|
|
|
|
# Block shapes — use auto-tuned config if available, else defaults
|
|
BLOCK_M = batch_info.max_len
|
|
config = get_lora_expand_config(
|
|
K=OUTPUT_DIM, R=MAX_RANK, num_slices=num_slices, chunk_size=BLOCK_M
|
|
)
|
|
BLOCK_K = config["BLOCK_K"]
|
|
BLOCK_N = config["BLOCK_N"]
|
|
|
|
num_segments = batch_info.num_segments
|
|
segment_grid = (
|
|
batch_info.weight_indices.shape[0]
|
|
if batch_info.use_cuda_graph
|
|
else num_segments
|
|
)
|
|
|
|
grid = (
|
|
triton.cdiv(max_slice_size, BLOCK_N),
|
|
num_slices, # number of slices in the input/output
|
|
segment_grid,
|
|
)
|
|
|
|
if base_output is None:
|
|
output = torch.zeros((M, OUTPUT_DIM), device=x.device, dtype=x.dtype)
|
|
else:
|
|
output = base_output
|
|
|
|
# Optional launch params from tuned config
|
|
extra_kwargs = {}
|
|
if "num_warps" in config:
|
|
extra_kwargs["num_warps"] = config["num_warps"]
|
|
if "num_stages" in config:
|
|
extra_kwargs["num_stages"] = config["num_stages"]
|
|
if "maxnreg" in config:
|
|
extra_kwargs["maxnreg"] = config["maxnreg"]
|
|
|
|
_chunked_lora_expand_kernel[grid](
|
|
x=x,
|
|
weights=weights,
|
|
output=output,
|
|
output_stride_0=output.stride(0),
|
|
output_stride_1=output.stride(1),
|
|
seg_indptr=batch_info.seg_indptr,
|
|
weight_indices=batch_info.weight_indices,
|
|
lora_ranks=batch_info.lora_ranks,
|
|
permutation=batch_info.permutation,
|
|
num_segs=segment_grid,
|
|
scalings=batch_info.scalings,
|
|
slice_offsets=slice_offsets,
|
|
# constants
|
|
NUM_SLICES=num_slices,
|
|
OUTPUT_DIM=OUTPUT_DIM,
|
|
MAX_RANK=MAX_RANK,
|
|
BLOCK_M=BLOCK_M,
|
|
BLOCK_N=BLOCK_N,
|
|
BLOCK_K=BLOCK_K,
|
|
**extra_kwargs,
|
|
)
|
|
|
|
return output
|